#include <io.h>                         // open() write() close()
#include <fcntl.h>                      // access for open()
#include <sys/stat.h>                   // create mode for open()
#include "..\intel\timestmp.h"          // class PentiumClock
#include "..\integrat\buffer.h"         // class BufferBlock, class BufferQueue
#include "..\integrat\flat.h"
#include "..\integrat\dump.h"
#include "..\intel\byteswap.h"
#include "..\integrat\trace.h"



// global variable
//
BufferQueue *all_buffers = NULL;



// constructor
//
// notes:
//      alignment takes care of start address of entire block
//      of all elements, not start address of 2+ elements, which
//      should be taken care of by element_size
//
BufferQueue::BufferQueue()
{
    // allocate memory for the control blocks
    //
    // they can be in low memory because they
    // only add up to about 100K bytes when we
    // have 62M bytes of 48-byte buffers to manage
    //
    total_blocks = SW_RINGS_WANTED;
    flat_free = (Bit8 *) align( RX_BUFFER_ALIGNMENT, flat_free );
    start_buffers = flat_free;
    start = new BufferBlock[ total_blocks ];
    end = start + total_blocks;

    // link all the elements together
    //
    unsigned count;
    BufferBlock *pointer = start;
    for( count=0; count<total_blocks-1; count++ )
    {
#if BUFFER_INIT_VERBOSE
        printf( "initializing block at 0x%p which controls buffer at flat 0x%8.8lX\n", pointer, pointer->buffers );
#endif

        pointer->next = (BufferBlock *) (pointer + 1);
        pointer++;
    }
    pointer->next = 0;

    // nothing is on the filled list
    //
    filled_head = 0;
    filled_tail = 0;
    filled_block_count = 0;
    filled_record_count = 0;

    // everything is on the free list
    //
    free_head = (BufferBlock *) start;

    // allocate() has never failed
    //
    allocate_failure_count = 0;
    prev_allocate_failure_count = 0;

    // dump file is not open
    //
    dump_handle = -1;
    dump_filename = NULL;
}



// should only be called by debug routines
// because it must iterate through the filled list
//
// index is zero-based
//
BufferBlock *BufferQueue::read( Bit16 cardnum, unsigned index )
{
    BufferBlock *pointer;
    unsigned count = 0;

    for( pointer=filled_head; pointer; pointer=pointer->next )
        if( pointer->cardnum == cardnum )
            if( count < index )
                count++;
            else
                break;

    return pointer;
}



// should only be called by debug routines
// because it doesn't disable interrupts and
// it must iterate through the entire
// free list, because we chose not to maintain
// counters during list manipulation operations,
// because we wanted those to be fast
//
Bit32 BufferQueue::amount_free( Bit32 &block_count )
{
    BufferBlock *block;
    Bit32 count = 0;
    Bit32 iterations;

    for( block=free_head,iterations=0; block && iterations<SW_RINGS; block=block->next,iterations++ )
    {
#if BUFFER_AMOUNT_FREE_VERBOSE
        printf( "BufferQueue::amount_free() traversing element at 0x%p\n", block );
#endif
        count += COMPL_RING_ENTRIES;
    }

    if( block )
        printf( "error!  BufferQueue 0x%p free list is probably circular!\n", this );

    block_count = iterations;
    return count;
}



// should only be called by debug routines
// because it must iterate through the entire
// filled list, because we chose not to maintain
// counters during list manipulation operations,
// because we wanted those to be fast
//
Bit32 BufferQueue::amount_filled( Bit16 cardnum, Bit32 &block_count )
{
    BufferBlock *block;
    Bit32 count = 0;
    Bit32 iterations;

    block_count = 0;
    for( block=filled_head,iterations=0; block && iterations<SW_RINGS; block=block->next,iterations++ )
    {
        if( block->cardnum == cardnum )
        {
#if BUFFER_AMOUNT_FILLED_VERBOSE
            printf( "BufferQueue::amount_filled() traversing element at 0x%p\n", block );
#endif
            count += block->used;
            block_count++;
        }
    }

    if( block )
        printf( "error!  BufferQueue 0x%p filled list is probably circular!\n", this );

    return count;
}



// make sure none of the buffer control blocks
// have become corrupted
//
// returns a nonzero value if an error is detected
//
// should only be called by debug routines
// because it must iterate through the entire
// filled list, and because it disables interrupts
//
// assumes that this BufferQueue was the first
// thing allocated out of locked memory since
// the program started
//
int BufferQueue::check( void )
{
    BufferBlock *block;
    Bit8 *buffers;
    unsigned nil_count = 0;
    unsigned count;
    BufferBlock *nil_ptrs[ QUEUE_MAX_NIL_PTRS ];
    int result = 0;
    Bit32 buffer_count;
    Bit32 block_count;
    Bit32 missing_buffers_calculated;
    Bit32 missing_blocks_calculated;
#if BUFFER_ALLOCATE_NONNIL_NEXT
    Bit32 missing_blocks_seen;
#endif

    if( this == 0 )
    {
        printf( "BufferQueue::check() called for unallocated queue!\n" );
        return ERROR_QUEUE_DEALLOCATED;
    }

    if( start == 0 )
    {
        printf( "BufferQueue::check() failed; start address is zero!\n" );
        return ERROR_QUEUE_STARTS_AT_ZERO;
    }

    // disable interrupts while we are iterating through
    // the queue, so it is not in flux, which can cause
    // momentary inconsistencies
    //
    // note 1: it is VERY important for the body of this
    // loop to not return directly to caller without
    // re-enabling interrupts
    //
    // note 2: printf() may enable interrupts internally,
    // so do not call it inside this loop if you intend
    // the loop to continue (better yet, don't do any I/O)
    //
    disable();

    missing_buffers_calculated = BUFFER_COUNT;
    missing_blocks_calculated = SW_RINGS;
    buffer_count = amount_filled( block_count );
    missing_buffers_calculated -= buffer_count;
    missing_blocks_calculated -= block_count;
    buffer_count = amount_free( block_count );
    missing_buffers_calculated -= buffer_count;
    missing_blocks_calculated -= block_count;
    if( missing_buffers_calculated > QUEUE_MAX_MISSING_BUFS
    ||  missing_blocks_calculated  > QUEUE_MAX_MISSING_BLKS )
    {
        printf( "BufferQueue::check() failed: %d/%d of %d/%d buffers/blocks are missing; should be no more than %d/%d\n",
            missing_buffers_calculated,
            missing_blocks_calculated,
            BUFFER_COUNT,
            SW_RINGS,
            QUEUE_MAX_MISSING_BUFS,
            QUEUE_MAX_MISSING_BLKS
        );

        enable();
        return ERROR_QUEUE_TOO_MANY_MISSING;
    }

#if BUFFER_ALLOCATE_NONNIL_NEXT
    missing_blocks_seen = 0;
#endif
    for( block=start,buffers=start_buffers; block<end; block++,buffers+=COMPL_RING_ENTRIES*BUFFER_SIZE )
    {
        if( block->buffers != buffers )
        {
            printf( "BufferQueue::check() failed at block 0x%p; should be"
                " pointing at flat address 0x%8.8lX but instead is 0x%8.8lX\n",
                block, buffers, block->buffers
            );

            result = ERROR_QUEUE_WRONG_FLAT_ADDR;
            break;
        }

        if( block->used > COMPL_RING_ENTRIES )
        {
            printf( "BufferQueue::check() failed at block 0x%p; should have"
                " used buffer count between 0 and %d, but instead has %d\n",
                block, COMPL_RING_ENTRIES, block->used
            );

            result = ERROR_QUEUE_WRONG_USED_COUNT;
            break;
        }

        if( !block->next )
        {
            nil_count++;
            if( nil_count <= QUEUE_MAX_NIL_PTRS )
                nil_ptrs[ nil_count - 1 ] = block;
            else
            {
                printf( "BufferQueue::check() failed at block 0x%p; next pointer"
                   " is nil, but %d nil pointers have already been seen at",
                   block, QUEUE_MAX_NIL_PTRS
                );
                for( count=0; count<nil_count-1; count++ )
                    printf( " 0x%p", nil_ptrs[count] );
                printf( "\n" );

                result = ERROR_QUEUE_TOO_MANY_NIL_PTRS;
                break;
            }
        }
#if BUFFER_ALLOCATE_NONNIL_NEXT
        else if( block->next == (BufferBlock *) 1 )
        {
            missing_blocks_seen++;
        }
#endif
        else if( !belongs( block->next ) )
        {
            printf( "BufferQueue::check() failed at block 0x%p; next pointer"
               " is 0x%p, not at an even boundary inside block array bounds of 0x%p, 0x%p\n",
               block, block->next, start, end
            );

            result = ERROR_QUEUE_PTR_OUT_OF_BOUNDS;
            break;
        }
    }   // end for all blocks

#if BUFFER_ALLOCATE_NONNIL_NEXT
    if( missing_blocks_seen != missing_blocks_calculated )
    {
        printf( "BufferQueue::check() failed; %d blocks were missing from the"
            " free and filled lists, but %d were tagged as allocated!\n",
            missing_blocks_calculated,
            missing_blocks_seen
        );

        result = ERROR_QUEUE_BAD_ALLOC_COUNT;
    }
#endif

#if !BUFFER_CHECK_DUMPS_WHEN_OK
    if( result != 0 )
    {
#endif

#if WAIT_FOR_KEYSTROKE
        printf( "\npress a key to see dump of all BufferBlock's\n" );
        getch();
        getch();
#endif
//        hex_dump_char( start, SW_RINGS*sizeof(BufferBlock), sizeof(BufferBlock) );
        hex_dump_char( block, sizeof(BufferBlock), sizeof(BufferBlock) );
#if !BUFFER_CHECK_DUMPS_WHEN_OK
    }
#endif

    // re-enable interrupts
    //
    enable();

    return result;
}



// assumes that dump filename is memory that will not change
// since we remember a pointer to it
//
int BufferQueue::dump_open( const char *const filename )
{
    // make sure the file is not already open (sequence error)
    //
    if( dump_handle >= 0 )
        return ERROR_QUEUE_FILE_ALREADY_OPEN;

    // create file with r/w directory attributes if it doesn't exist
    // truncate file if it already exists
    // writes to files will be binary (no newline translation)
    // during exectuion, we need only write access
    //
    dump_handle = open( filename, O_CREAT|O_TRUNC|O_BINARY|O_WRONLY, S_IWRITE|S_IREAD );
    if( dump_handle < 0 )
    {
        printf( "error %d opening file '%s' for writing\n", errno, filename );
        return ERROR_QUEUE_FILE_CANNOT_OPEN;
    }

    dump_filename = filename;
    return 0;
}



int BufferQueue::dump_close( void )
{
    int result;

    // make sure the file is already open (sequence error)
    //
    if( dump_handle < 0 )
        return ERROR_QUEUE_FILE_NOT_OPEN;

    result = close( dump_handle );
    if( result < 0 )
    {
        printf( "error %d closing dump file '%s'\n", errno, dump_filename );
        result = WARNING_QUEUE_FILE_CANNOT_CLOSE;
    }
    else
    {
        result = 0;
        dump_handle = -1;
        dump_filename = NULL;
    }

    return result;
}



int BufferQueue::dump_one( BufferBlock *block )
{
    Bit32 count;
    Bit32 temp;
    int result;
    static struct                   // must align with struct traceblock in trace.h
    {
        Bit32 used;
        Bit16 cardnum;
        Bit8 cardindex;
        Bit8 padding[ SECTOR_SIZE - sizeof(Bit8) - sizeof(Bit16) - sizeof(Bit32) ];
#if BUFFER_DUMP_COPIES_DATA
        Bit8 data[ COMPL_RING_ENTRIES ][ BUFFER_SIZE ];
#endif
    } shadow;

    // make sure the file is already open (sequence error)
    //
    if( dump_handle < 0 )
        return ERROR_QUEUE_FILE_NOT_OPEN;

    // copy the time stamp and card number to the front
    // of the buffer, so we can write everything out to
    // disk at one time
    //
    shadow.used = htonl( block->used );
    shadow.cardnum = htons( block->cardnum );
    shadow.cardindex = block->cardindex;

    // zero out the unused buffers
    //
    if( block->used < COMPL_RING_ENTRIES )
        memset( block->buffers+(block->used*BUFFER_SIZE), 0, BUFFER_SIZE*(COMPL_RING_ENTRIES - block->used) );

    // for all records in the BufferBlock
    //
    for( count=0; count<block->used; count++ )
    {
        tracerec *packet = (tracerec *) (block->buffers + count*BUFFER_SIZE);

        // swap the bytes and words of the timestamp
        //
        temp = htonl( packet->when.low );
        packet->when.low  = htonl( packet->when.high );
        packet->when.high = temp;

        // swap the bytes of the ATM header
        //
        packet->atm.whole = htonl( packet->atm.whole );

        // the rest of the data is IP and TCP headers,
        // and was always in network byte order!
        //
    }

#if BUFFER_DUMP_COPIES_DATA
    // copy the packet next to the header so we can do a single disk write
    //
    memcpy( &shadow.data, block->buffers, BUFFER_SIZE*COMPL_RING_ENTRIES );
#endif

    // write the packet header (and data if it was copied) to disk
    //
    result = write( dump_handle, &shadow, sizeof(shadow) );
    if( result < 0 )
    {
        printf( "error %d writing block header at 0x%8.8lX to file '%s'\n", errno, &shadow, dump_filename );
        return ERROR_QUEUE_FILE_CANNOT_WRITE;
    }

#if !BUFFER_DUMP_COPIES_DATA
    // write the packet data to disk
    //
    // note that we write COMPL_RING_ENTRIES buffers
    // even when not all of them are filled, to simplify
    // code that reads this file
    //
    // I think this (unfortunately) waits for completion
    // of previous write()
    //
    result = write( dump_handle, block->buffers, BUFFER_SIZE*COMPL_RING_ENTRIES );
    if( result < 0 )
    {
        printf( "error %d writing block at 0x%8.8lX with %d packets used to file '%s'\n", errno, block->buffers, block->used, dump_filename );
        return ERROR_QUEUE_FILE_CANNOT_WRITE;
    }
#endif

    return 0;
}



// write the packets and headers in a buffer queue to disk
//
// warning: this empties the filled list to the free list
//
// when fewer than COMPL_RING_ENTRIES are used in a block,
// we still write a whole block to disk, to simplify programs
// that process the file
//
//
//
// warning: if you do not have DOS=HIGH in C:\CONFIG.SYS
// you will have to do gate_a20(1) after calling this function
// because DOS did us a "favor" when we did file I/O
//
// this would be a BIG problem to do in mainline
// code, i.e. while packets are still coming in,
// because a foreground DOS file I/O call would
// cause our interrupt handler to fail!
//
// and no, we cannot call gate_a20() in an interrupt
// handler, because it loops up to 127K times while
// waiting for keyboard controller to respond!!!
//
//
//
// warning: if you didn't disable caching above 64MB in BIOS
// setup utility, machine may lock up when this code touches
// above that line
//
// this may be a problem with Borland's write() function
//
// disabling caching above 64MB doesn't affect speed of
// writing to disk... still horribly slow at 6.5 seconds to
// write one block of 1 MB... I gotta write code to go
// direct to EIDE controller sometime soon!
//
int BufferQueue::dump( const char *const filename )
{
    int result, result2;
    BufferBlock *block;
#if BUFFER_DUMP_PARANOIA
    Bit32 infinite_loop_protection = 0;
#endif
#if ESSENTIAL_TIME_STAMPS
    PentiumClock start, end;
    double timing;
    Bit32 total;
#endif

#if ESSENTIAL_TIME_STAMPS
    amount_filled( total );
    start.set();
#endif

    result = dump_open( filename );
    if( result != 0 )
    {
        return ERROR_QUEUE_FILE_CANNOT_OPEN;
    }

    while( 1 )
    {
        // stop if user pressed a key
        //
        if( kbhit() )
        {
            printf( "stopping because of keypress\n" );
            result = WARNING_QUEUE_FILE_KEYPRESS;
            break;
        }

        // take a block of buffers off the filled list
        //
        result = read( &block );
        if( result != 0 )
        {
            if( result == ERROR_QUEUE_NOT_FILLED )
                result = 0;
            else
                printf( "error %d reading a block of packets from queue to dump to file '%s'\n", result, filename );

            break;
        }

        result = dump_one( block );
        if( result != 0 )
            break;

#if BUFFER_DUMP_TICKS
        fprintf( stderr, "." );
#endif

#if BUFFER_DUMP_TRACE
        fprintf( stderr, "wrote block of %d/%d packets/bytes from log/phy 0x%8.8lX/0x%8.8lX\n",
            block->used, BUFFER_SIZE*COMPL_RING_ENTRIES,
            block->buffers, flat_to_physical(block->buffers)
        );
#endif

        // put the block of buffers onto the free list
        // so it is not lost forever
        //
        result = free( block );
        if( result != 0 )
        {
            printf( "error %d freeing buffer block 0x%p after writing to file '%s'\n", result, block, filename );
            break;
        }

#if BUFFER_DUMP_PARANOIA
        infinite_loop_protection++;
        if( infinite_loop_protection > total )
        {
            printf( "Buffer::dump() internal error!  >%d buffer blocks returned\n", total );
        }
#endif
    }

    result2 = dump_close();
    if( result2 && !result )
        result = result2;

#if ESSENTIAL_TIME_STAMPS
    end.set();
    total *= COMPL_RING_ENTRIES;
    timing = PentiumClock( end - start ).get_seconds();
    printf( "BufferQueue %p dump took %g S total\n", this, timing );
    timing = timing * 1000 / total;
    printf( "BufferQueue %p dump took %g mS per %d byte buffer\n", this, timing, BUFFER_SIZE );
#endif

    return result;
}
