#include <stdio.h>                  // printf()
#include <fcntl.h>                  // O_CREAT etc.
#include <io.h>                     // open() close() read() write()
#include <stdlib.h>                 // atoi()
#include <stdarg.h>                 // va_start() va_end() va_list
#include <string.h>                 // stricmp()
#include <mem.h>                    // memset()
#include <errno.h>                  // errno _sys_errlist
#include <conio.h>                  // kbhit()
#include "..\intel\timestmp.h"
#include "..\wattcp\tcp.h"
#include "..\integrat\flat.h"       // flat_alloc()

void init_trace(int level);



#define FORCE_CACHING           FALSE
#define USE_NET_BUFFER          TRUE
#define SHOW_REAL_MEM_USAGE     FALSE
#define VERBOSE_TCP_FAILURES    TRUE
#define VERBOSE_FLAT_INIT       TRUE



#define CONTROL_WAIT            100    // measured in seconds
#define FILENAME_SIZE           256
#define MODE_SIZE               100
#define NET_BUFFER_SIZE         (11*1024*1024)
#define APP_BUFFER_SIZE         (10*1024*1024)     // want an even multiple of 512 bytes for fast disk I/O



#define TRUE                    1
#define FALSE                   0



char *app_buffer;
#if USE_NET_BUFFER
    char *net_buffer;
#endif



// wait for data to be available
// or socket to close
//
void wait_for_data( tcp_Socket *sock )
{
    while( tcp_tick( sock ) && !sock_dataready( sock ) )
        ;
}



int cleanup( int retcode, tcp_Socket *sock, const char *format, ... )
{
    va_list argptr;

    va_start( argptr, format );
    vprintf( format, argptr );
    va_end( argptr );

    // signal that we want socket closed
    //
    sock_close( sock );

    // wait for socket to actually close
    //
    // does this involve 3 minute TIME_WAIT???
    //
    while( tcp_tick( sock ) )
        ;

    // return to OS
    //
    return retcode;
}



int sock_to_file( tcp_Socket *sock, const char *filename, long *bytes, PentiumClock *disk_io_time )
{
    int result, result2;
    int file_handle;
    PentiumClock start, end;

    // log to the console
    //
    printf( "copying TCP/IP stream to file '%s'\n", filename );

    // no bytes RX yet
    //
    *bytes = 0;

    // remember when this disk I/O started
    //
    // note: for clarity, will not be commented
    // on subsequent disk I/O
    //
    start.set();

    // open up the file locally
    //
    file_handle = open( filename, O_CREAT|O_TRUNC|O_BINARY|O_WRONLY );

    // get the disk I/O end time, calculate the delta,
    // and add it to the caller's total
    //
    // note: for clarity, will not be commented
    // on subsequent disk I/O
    //
    end.set();
    *disk_io_time += end - start;

    // check the disk I/O result
    //
    if( file_handle == -1 )
    {
        return cleanup( 10, sock, "cannot create file '%s'\n", filename );
    }

#if USE_NET_BUFFER
    // tell socket to use my buffer, not his, for receive
    //
    if( !sock_set_rx_buf( sock, net_buffer, NET_BUFFER_SIZE ) )
        return cleanup( 12, sock, "error setting socket to have %d byte receive buffer\n", NET_BUFFER_SIZE );
#endif

    // change socket to binary mode so all bytes get copied verbatin
    //
    sock_mode( sock, TCP_MODE_BINARY );

    // while client hasn't closed socket...
    //
    while( tcp_tick( sock ) )
    {
        // stop if the user pressed a key
        //
        if( kbhit() )
            return cleanup( 11, sock, "\nhalted by keypress\n" );

        // read up to one buffer's worth of data
        //
        // note: cannot use sock_fastread() here because
        // we want our disk write to be an even multiple
        // of 512 bytes so it will be fast
        //
        result = sock_read( sock, app_buffer, APP_BUFFER_SIZE );

        // if we got anything...
        //
        if( result > 0 )
        {
            // write it to disk file
            //
            start.set();
            result2 = write( file_handle, app_buffer, (unsigned)result );
            end.set();
            *disk_io_time += end - start;
            if( result2 != (unsigned)result )
            {
                close( file_handle );
                return cleanup( 12, sock, "could only write %d of %d bytes to file '%s'\n", result2, result, filename );
            }

            // keep track of how many bytes we have done
            //
            *bytes += result;
        }
    }

    // close the disk file
    //
    start.set();
    close( file_handle );
    end.set();
    *disk_io_time += end - start;

    // exit to OS successfully
    //
    return 0;
}



int file_to_sock( tcp_Socket *sock, const char *filename, long *bytes, PentiumClock *disk_io_time )
{
    int result, result2;
    int file_handle;
    PentiumClock start, end;

    // log to the console
    //
    printf( "copying file '%s' to TCP/IP stream\n", filename );

    // no bytes TX yet
    //
    *bytes = 0;

    // remember when this disk I/O started
    //
    // note: for clarity, will not be commented
    // on subsequent disk I/O
    //
    start.set();

    // open up the file locally
    //
    file_handle = open( filename, O_BINARY|O_RDONLY );

    // get the disk I/O end time, calculate the delta,
    // and add it to the caller's total
    //
    // note: for clarity, will not be commented
    // on subsequent disk I/O
    //
    end.set();
    *disk_io_time += end - start;

    // check the disk I/O result
    //
    if( file_handle == -1 )
    {
        return cleanup( 20, sock, "error %d '%s' opening existing file '%s'\n", errno, _sys_errlist[errno], filename );
    }

#if USE_NET_BUFFER
    // tell socket to use my buffer, not his, for transmit
    //
    if( !sock_set_tx_buf( sock, net_buffer, NET_BUFFER_SIZE ) )
        return cleanup( 12, sock, "error setting socket to have %d byte transmit buffer\n", NET_BUFFER_SIZE );
#endif

    // change socket to binary mode so all bytes get copied verbatin
    //
    sock_mode( sock, TCP_MODE_BINARY );

    // while client hasn't closed socket
    // and we haven't reached end of file...
    //
    result = INT_MAX;           // not at EOF yet
    while( tcp_tick( sock ) && result>0 )
    {
        // stop if the user pressed a key
        //
        if( kbhit() )
            return cleanup( 22, sock, "\nhalted by keypress\n" );

        // read up to one buffer's worth of data
        //
        // note: error return code -1 is treated like EOF retcode 0
        //
        start.set();
        result = read( file_handle, app_buffer, APP_BUFFER_SIZE );
        end.set();
        *disk_io_time += end - start;

        // if we got anything...
        //
        if( result > 0 )
        {
            // write it to socket
            //
            result2 = sock_write( sock, app_buffer, result );
            if( result2 != result )
            {
                close( file_handle );
                return cleanup( 23, sock, "could only write %d of %d bytes to socket\n", result2, result );
            }

            // keep track of how many bytes we have done
            //
            *bytes += result;
        }
    }

    // close the disk file
    //
    start.set();
    close( file_handle );
    end.set();
    *disk_io_time += end - start;

    // close the socket ourselves and
    // exit to OS successfully
    //
    return cleanup( 0, sock, "" );
}



int daemon( tcp_Socket *sock, unsigned port )
{
    int result;
    word count;
    char filename[ FILENAME_SIZE + 1 ];
    char mode_string[ MODE_SIZE + 1 ];
    int mode_is_upload;
    PentiumClock start, overall, disk, network;
    long bytes;

    // listen for a connection
    //
    result = tcp_listen( sock, port, 0, 0, NULL, 0 );

    // wait for a connection to come in
    //
    sock_wait_established( sock, CONTROL_WAIT, NULL, &result );

    // remember when data transfer started
    //
    start.set();

    // change to ASCII mode because we don't want to be told
    // that any data is ready unless it is a complete line
    //
    sock_mode( sock, TCP_MODE_ASCII );

    wait_for_data( sock );

    // read the first line as command
    //
    count = sock_gets( sock, mode_string, MODE_SIZE );
    if( count == 0 )
    {
        return cleanup( 30, sock, "mode not provided!\n" );
    }
    mode_is_upload = stricmp( mode_string, "RETR" );

    wait_for_data( sock );

    // read the second line as a filename
    //
    count = sock_gets( sock, filename, FILENAME_SIZE );
    if( count == 0 )
    {
        return cleanup( 31, sock, "filename not provided!\n" );
    }

    disk.low = 0;
    disk.high = 0;
    if( mode_is_upload )
    {
        // suck up the last trailing newline
        //
        wait_for_data( sock );

        result = sock_to_file( sock, filename, &bytes, &disk );
    }
    else
        result = file_to_sock( sock, filename, &bytes, &disk );

    // calculate how much time has elapsed
    //
    overall.set();
    overall -= start;
    network = overall - disk;

    // inform the user how much we got and how fast
    //
    printf( "%.3f S overall = %.3f disk + %.3f network\n%ld bytes, %.3f network KB/S\n",
       overall.get_seconds(), disk.get_seconds(), network.get_seconds(),
       bytes, ((double)bytes/1000.0)/network.get_seconds()
    );

    return result;

sock_err:
    printf( "got error %d from sock_wait_established\n", result );
    return 999;
}



int client( tcp_Socket *sock, unsigned port, char *hostname, const char *mode_string, char *local_filename, char *remote_filename )
{
    int result;
    int mode_is_upload;
    longword hostaddr;
    char *command;
    PentiumClock start, overall, disk, network;
    long bytes;

    // resolve the host name to an address
    //
    hostaddr = resolve( hostname );
    if( !hostaddr )
    {
        printf( "unable to resolve host name '%s' to numerical IP address\n", hostname );
        return 40;
    }

    // convert the mode into a command
    //
    if( !stricmp( mode_string, "u" ) )
    {
        mode_is_upload = TRUE;
        command = "SAVE";
    }
    else if( !stricmp( mode_string, "d" ) )
    {
        mode_is_upload = FALSE;
        command = "RETR";
    }
    else
    {
        printf( "invalid file transfer mode '%s'\n", mode_string );
        return 41;
    }

    // initiate a connection
    //
    result = tcp_open( sock, 0, hostaddr, port, NULL );

    // wait for our connection to finish
    //
    sock_wait_established( sock, CONTROL_WAIT, NULL, &result );

    // remember when data transfer started
    //
    start.set();

    // change to ASCII mode because we will be writing strings
    //
    sock_mode( sock, TCP_MODE_ASCII );

    // write the command as the first line
    //
    sock_puts( sock, command );

    // write the filename as the second line
    //
    sock_puts( sock, remote_filename );

    disk.low = 0;
    disk.high = 0;
    if( mode_is_upload )
        result = file_to_sock( sock, local_filename, &bytes, &disk );
    else
        result = sock_to_file( sock, local_filename, &bytes, &disk );

    // calculate how much time has elapsed
    //
    overall.set();
    overall -= start;
    network = overall - disk;

    // inform the user how much we sent and how fast
    //
    printf( "%.3f S overall = %.3f disk + %.3f network\n%ld bytes, %.3f network KB/S\n",
       overall.get_seconds(), disk.get_seconds(), network.get_seconds(),
       bytes, ((double)bytes/1000.0)/network.get_seconds()
    );

    return result;

sock_err:
    printf( "got error %d from sock_wait_established\n", result );
    return 999;
}



int main( int argc, char *argv[] )
{
    const char *program = argv[0];
    unsigned control_port;
    int result;
    tcp_Socket control_socket;

    if( argc!=2 && argc!=5 && argc!=6 )
    {
        printf( "daemon:          %s tcp-port-number\n", program );
        printf( "client   upload: %s tcp-port-number hostname u local-filename {remote-filename}\n", program );
        printf( "client download: %s tcp-port-number hostname d local-filename {remote-filename}\n", program );
        return 1;
    }

    control_port = atoi( argv[ 1 ] );

    // initialize TCP/IP library
    //
    sock_init();
#if VERBOSE_TCP_FAILURES
    tcp_set_debug_state( 1 );
#endif
#if SHOW_REAL_MEM_USAGE
    system( "mem /d | find \"TCPCOPYD\"" );
#endif

//   tcp_set_debug_state(1);  /* Display flags for each packet */
//   tcp_set_debug_state(2);  /* Display full trace data */
//   init_trace(2);  /* Nevil's tcp packet trace - full */
//   init_trace(1);  /* Nevil's tcp packet trace - times only */

    // initialize access to our own area of flat memory
    //
    flat_init( VERBOSE_FLAT_INIT );

    // allocate buffers
    //
    app_buffer = (char *) flat_alloc( APP_BUFFER_SIZE );
    if( !app_buffer )
    {
        printf( "error %d allocating %d bytes for application buffer\n", errno, APP_BUFFER_SIZE );
        return 50;
    }

#if USE_NET_BUFFER
    net_buffer = (char *) flat_alloc( NET_BUFFER_SIZE );
    if( !net_buffer )
    {
        printf( "error %d allocating %d bytes for network buffer\n", errno, NET_BUFFER_SIZE );
        return 51;
    }
#endif

#if FORCE_CACHING
    // touch all words of all buffers to force caching to a known state
    //
    // does this make timings more consistent?
    //
    memset( app_buffer, 0, APP_BUFFER_SIZE );
#if USE_NET_BUFFER
    memset( net_buffer, 0, NET_BUFFER_SIZE );
#endif
#endif

    // set socket to known state
    //
    memset( &control_socket, 0, sizeof(control_socket) );

    if( argc == 2 )
    {
        result = daemon( &control_socket, control_port );
    }
    else
    {
        result = client( &control_socket, control_port, argv[2], argv[3], argv[4], argc==6?argv[5]:argv[4] );
    }

    return result;
}



// written 7/17/96 by Joel Apisdorf
//
